/***************************************************************************
                          hasegawa85model.cpp  -  description
                             -------------------
    begin                : Thu Feb 20 2003
    copyright            : (C) 2003 by Vinh Le Sy and Arndt von Haeseler
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <math.h>
#include <iostream>

#include "hky85m.h"
#include "brent.h"
#include "opturtree.h"
#include "outstream.h"

//the constructor function of this class
HKY85M::HKY85M () {
}

//--------------------------------------------------------------------
//compute the substitution rate per unit time for this model
double HKY85M::cmpSubRate () {
  double tmpSubRate_ = 0.0;
  double purineFrq_ = stateFrqArr_[BASE_A] + stateFrqArr_[BASE_G];
  double pyriFrq_ = stateFrqArr_[BASE_C] + stateFrqArr_[BASE_T];
  //for BASE_ A
  tmpSubRate_ += stateFrqArr_[BASE_A] * (pyriFrq_ + tsTvRatio_ * stateFrqArr_[BASE_G]);
  //for BASE_ C
  tmpSubRate_ += stateFrqArr_[BASE_C] * (purineFrq_ + tsTvRatio_ * stateFrqArr_[BASE_T]);
  //for BASE_ G
  tmpSubRate_ += stateFrqArr_[BASE_G] * (pyriFrq_ + tsTvRatio_ * stateFrqArr_[BASE_A]);
  //for BASE_ T
  tmpSubRate_ += stateFrqArr_[BASE_T] * (purineFrq_ + tsTvRatio_ * stateFrqArr_[BASE_C]);
  return tmpSubRate_;
}

//--------------------------------------------------------------------
//all things are inited here
void HKY85M::init () {
  subRate_ = cmpSubRate ();
}

//--------------------------------------------------------------------
//all things are inited here
void HKY85M::reCmpSubRate () {
  subRate_ = cmpSubRate ();
}

//--------------------------------------------------------------------
/*
cmp the probability of changing from nucleotide stateNo1 into nucleotide stateNo2
after a period of brLen / subRate_
*/
double HKY85M::cmpProbChange (const int stateNo1, const int stateNo2, const double brLen) {
  double prob_;
  double time_ = brLen  / subRate_;
  double totalState_;
  if (stateNo2 == BASE_A || stateNo2 == BASE_G)
    totalState_ = stateFrqArr_[BASE_A] + stateFrqArr_[BASE_G];
  else
    totalState_ = stateFrqArr_[BASE_C] + stateFrqArr_[BASE_T];	

  double ax_ = 1.0 + totalState_ * (tsTvRatio_ - 1.0);	

  if (stateNo1 == stateNo2) {
    prob_ = stateFrqArr_[stateNo2];
    prob_ += stateFrqArr_[stateNo2] * ( (1.0 / totalState_) - 1.0) * exp (-time_);
    prob_ += ( (totalState_ - stateFrqArr_[stateNo2]) / totalState_) * exp (-time_ * ax_);
  } else {
    //count probability of ts (purine -> purine, or pyridimine -> pyridimine)
    if ( stateNo1 + stateNo2 == TS_SIGN ) {
      prob_ = stateFrqArr_[stateNo2];
      prob_ += stateFrqArr_[stateNo2] * ( (1.0 / totalState_) - 1.0) * exp (-time_);
      prob_ -= (stateFrqArr_[stateNo2] / totalState_) * exp (-time_ * ax_);
    } else //count probability of tv (purine <-> pyridimine)
      prob_ = stateFrqArr_[stateNo2] * (1.0 - exp (-time_) );
  }

  return prob_;
}



//--------------------------------------------------------------------
//cmp the vector probability of change from stateNo
void HKY85M::cmpProbChange (const int stateNo, const double brLen, DVec20 &probVec) {
  for (int stateNo2_ = 0; stateNo2_ < NUM_BASE; stateNo2_++)
    probVec[stateNo2_] = cmpProbChange (stateNo, stateNo2_, brLen);
}


//--------------------------------------------------------------------
/*
cmp the array of probability of changing from one nucleotide 
into one nucleotide after a period of brLen / subRate_
*/
void HKY85M::cmpProbChange (const double brLen, DMat20 &probMat) {
  for (int stateNo1_ = 0; stateNo1_ < NUM_BASE; stateNo1_++) {
    probMat[stateNo1_][stateNo1_] = cmpProbChange (stateNo1_, stateNo1_, brLen);
    for (int stateNo2_ = stateNo1_ + 1; stateNo2_ < NUM_BASE; stateNo2_++) {
      probMat[stateNo1_][stateNo2_] = cmpProbChange (stateNo1_, stateNo2_, brLen);
      probMat[stateNo2_][stateNo1_] = cmpProbChange (stateNo2_, stateNo1_, brLen);
    } //end for stateNo2
  } //end for stateNo1 
}


//--------------------------------------------------------------------
/*compute the log likelihood when given tsTvRatio,
this function is used to optimize the tsTvRatio of this model
*/
double HKY85M::cmpNegLogLiTsTvRatio (double tsTvRatio) {
  tsTvRatio_ = tsTvRatio;
  if (tsTvRatio_ < MIN_TS_TV_RATIO)
    tsTvRatio_ = MIN_TS_TV_RATIO;
  if (tsTvRatio_ > MAX_TS_TV_RATIO)
    tsTvRatio_ = MAX_TS_TV_RATIO;

  reCmpSubRate ();
  OptUrTree::cmpLiNd ();
  OptUrTree::opt (MAX_IT_UR_BR_PAM_OPT);
  double logLi_ = OptUrTree::getLogLi ();
  return -logLi_;
}

//--------------------------------------------------------------------
//optimize the tsTvRatio basing Brent method
void HKY85M::optPam () {
  double fx_, error_;
  std::cout <<"Optimizing transition/transversion ratio ..." << endl;
  Brent::turnOnOptedPam ();
  tsTvRatio_= Brent::optOneDim (MIN_TS_TV_RATIO, tsTvRatio_, MAX_TS_TV_RATIO,
                                cmpNegLogLiTsTvRatio, EPS_MODEL_PAM_ERROR, &fx_, &error_);  
  Brent::turnOffOptedPam ();

  reCmpSubRate ();
  OptUrTree::cmpLiNd ();

//  double logLi_ = OptUrTree::getLogLi ();
  std::cout.precision (10);
//  std::cout <<"Log likelihood: " << logLi_ << endl;

  std::cout.precision (5);
  std::cout << "Transition/transversion ratio: " << tsTvRatio_ / 2.0 << endl;
}


//--------------------------------------------------------------------
//release all memory of this class
void HKY85M::release () {
}

//--------------------------------------------------------------------
//the destructor function of this class
HKY85M::~HKY85M (){
  release ();
//  std::cout << "this is the destructor of HKY85M class " << endl;
}
